#include "ShmemLockedQueue.h"
#include <stdio.h>
#include "lock.h"

// Function name	: ShmemLockedQueue::ShmemLockedQueue
// Description	    : 
// Return type		: 
ShmemLockedQueue::ShmemLockedQueue()
{
	m_dwSize = 0;
	m_pBase = NULL;
	m_pBottom = NULL;
	m_hMapping = NULL;
	m_hQMutex = NULL;
	m_hMsgAvailableEvent = NULL;
	m_hQEmptyEvent = NULL;
}

// Function name	: ShmemLockedQueue::~ShmemLockedQueue
// Description	    : 
// Return type		: 
ShmemLockedQueue::~ShmemLockedQueue()
{
	if (m_hMapping != NULL)
	{
		if (m_pBottom != NULL)
			UnmapViewOfFile(m_pBottom);
		CloseHandle(m_hMapping);
	}
	if (m_hMsgAvailableEvent != NULL)
		CloseHandle(m_hMsgAvailableEvent);
	m_hMsgAvailableEvent = NULL;
}

// Function name	: ShmemLockedQueue::Init
// Description	    : 
// Return type		: bool 
// Argument         : char *name
// Argument         : unsigned long size
bool ShmemLockedQueue::Init(char *name, unsigned long size)
{
	bool bFirst = true;

	// Create a mapping from the page file
	m_hMapping = CreateFileMapping(
		INVALID_HANDLE_VALUE,
		NULL,
		PAGE_READWRITE,
		0, size,
		name);

	// If the mapping already exists then we are attaching to it, not creating it
	if (GetLastError() == ERROR_ALREADY_EXISTS)
		bFirst = false;

	if (m_hMapping == NULL)
		return false;

	// Map the file and save the pointer to the base of the mapped file
	m_pBottom = MapViewOfFile(
		m_hMapping,
		FILE_MAP_WRITE,
		0,0,
		size);

	if (m_pBottom == NULL)
	{
		CloseHandle(m_hMapping);
		m_hMapping = NULL;
		return false;
	}

	m_hQMutex = (LONG*)m_pBottom;
	m_hQEmptyEvent = &((LONG*)m_pBottom)[1];
	m_hMsgAvailableEvent = &((LONG*)m_pBottom)[2];

	m_pEnd = (LPBYTE)m_pBottom + size;
	m_pBase = (LPBYTE)m_pBottom + 3*sizeof(LONG);
	m_dwSize = size;
	// max message = total mapped space - head and tail pointer - one message header - 3 locks
	m_dwMaxMsgSize = size - (2*sizeof(ShmemLockedQueueHeader*)) - sizeof(ShmemLockedQueueHeader) - 3*sizeof(LONG);

	// If this process is creating the mapping, then set up the head and tail pointers
	if (bFirst)
	{
		((unsigned long*)m_pBase)[0] = 0;
		((unsigned long*)m_pBase)[1] = 2*sizeof(unsigned long);
		*m_hQMutex = 0;
		*m_hQEmptyEvent = 0;
		//*m_hMsgAvailableEvent = 0;
	}

	// Create or Attach to the synchronization handles for this queue
	char pszEventName[200];

	sprintf(pszEventName, "%s.event", name);

	m_hMsgAvailableEvent = CreateEvent(NULL, TRUE, FALSE, pszEventName);
	if (m_hMsgAvailableEvent == NULL)
	{
		CloseHandle(m_hMapping);
		CloseHandle(m_hQMutex);
		m_hMapping = NULL;
		m_hQMutex = NULL;
		return false;
	}

	return true;
}

// Function name	: ShmemLockedQueue::Insert
// Description	    : 
// Return type		: bool 
// Argument         : unsigned char *buffer
// Argument         : unsigned int length
// Argument         : int tag
// Argument         : int from
bool ShmemLockedQueue::Insert(unsigned char *buffer, unsigned int length, int tag, int from)
{
	ShmemLockedQueueHeader *pTail;

	lock(m_hQMutex);

	if (length > m_dwMaxMsgSize)
	{
		unlock(m_hQMutex);
		return false;
	}

	// Wait for a contiguous block large enough to hold the data 
	while ((unsigned long)m_pEnd - (unsigned long)m_pBase - ((unsigned long *)m_pBase)[1] - (unsigned long)sizeof(ShmemLockedQueueHeader) < length)
	{
		unlock(m_hQMutex);
		wait(m_hQEmptyEvent);
		lock(m_hQMutex);
	}

	// Read the tail pointer
	pTail = (ShmemLockedQueueHeader *)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
	// If the head pointer is 0, set it to the tail
	if ( *((unsigned long *)m_pBase) == 0 )
		*((unsigned long *)m_pBase) = (unsigned long)((LPBYTE)pTail - (LPBYTE)m_pBase);

	// Set the state and advance the tail offset
	pTail->state = SHMEM_Q_BEING_WRITTEN;
	// Advance the tail pointer
	((unsigned long *)m_pBase)[1] = (unsigned long)((unsigned long)pTail + length + sizeof(ShmemLockedQueueHeader) - (unsigned long)m_pBase);
	unlock(m_hQMutex);

	// Write the header
	pTail->tag = tag;
	pTail->from = from;
	pTail->length = length;
	pTail->next_offset = (sizeof(ShmemLockedQueueHeader) + length);
	
	// Copy the data
	memcpy((LPBYTE)pTail + sizeof(ShmemLockedQueueHeader), buffer, length);

	lock(m_hQMutex);
	// Signal data has arrived and release the mutex
	pTail->state = SHMEM_Q_AVAIL_FOR_READ;
	//setevent(m_hMsgAvailableEvent);
	SetEvent(m_hMsgAvailableEvent);
	resetevent(m_hQEmptyEvent);
	unlock(m_hQMutex);

	return true;
}

// Function name	: ShmemLockedQueue::RemoveNext
// Description	    : 
// Return type		: bool 
// Argument         : unsigned char *buffer
// Argument         : unsigned int *length
// Argument         : int *tag
// Argument         : int *from
bool ShmemLockedQueue::RemoveNext(unsigned char *buffer, unsigned int *length, int *tag, int *from)
{
	ShmemLockedQueueHeader *pTail, *pHead, *pMessage;

	// Get the next available entry in the queue
	while(true)
	{
		// Get the queue mutex
		lock(m_hQMutex);
		
		// Wait for the queue to not be empty
		while (((unsigned long*)m_pBase)[0] == 0)
		{
			unlock(m_hQMutex);
			//wait(m_hMsgAvailableEvent);
			if (WaitForSingleObject(m_hMsgAvailableEvent, INFINITE) != WAIT_OBJECT_0)
			{
				printf("Wait for MsgAvailableEvent failed\n");
				return false;
			}
			lock(m_hQMutex);
		}
		
		// Search the queue for the next available entry
		pMessage = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
		pTail = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
		while ((pMessage->state == SHMEM_Q_BEING_READ) && (pMessage < pTail))
			pMessage = (ShmemLockedQueueHeader*)((LPBYTE)pMessage + pMessage->next_offset);
		
		// If we haven't reached the tail and the element we are on is not currently being written, then successfully break out of this loop
		if ( (pMessage < pTail) && (pMessage->state != SHMEM_Q_BEING_WRITTEN) )
			break;

		// All messages are being read or the next message in order is not ready. I need to reset MsgAvailableEvent, wait for it to be signalled and then start over
		//resetevent(m_hMsgAvailableEvent);
		ResetEvent(m_hMsgAvailableEvent);
		unlock(m_hQMutex);
		//wait(m_hMsgAvailableEvent);
		if (WaitForSingleObject(m_hMsgAvailableEvent, INFINITE) != WAIT_OBJECT_0)
		{
			printf("Wait for MsgAvailableEvent failed\n");
			return false;
		}
	}

	// Check that the buffer provided is large enough to hold the data
	if (pMessage->length > *length)
	{
		printf("length %d > *length %d\n", pMessage->length, *length);
		ReleaseMutex(m_hQMutex);
		return false;
	}

	// Mark the message as being read
	pMessage->state = SHMEM_Q_BEING_READ;
	unlock(m_hQMutex);

	// Read the data from the message
	*tag = pMessage->tag;
	*from = pMessage->from;
	*length = pMessage->length;
	memcpy(buffer, (LPBYTE)pMessage + sizeof(ShmemLockedQueueHeader), pMessage->length);

	lock(m_hQMutex);
	// Mark the message as having been read
	pMessage->state = SHMEM_Q_READ;

	// Update the head and tail pointers of the queue
	pHead = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
	pTail = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
	// Advance the head pointer over all the read messages
	while ((pHead->state == SHMEM_Q_READ) && (pHead < pTail))
		pHead = (ShmemLockedQueueHeader*)((LPBYTE)pHead + pHead->next_offset);

	if (pHead >= pTail)
	{
		// When the head catches up to the tail, the queue is empty so reset the pointers and signal the queue empty
		((unsigned long*)m_pBase)[0] = 0;
		((unsigned long*)m_pBase)[1] = 2*sizeof(unsigned long);
		//resetevent(m_hMsgAvailableEvent);
		ResetEvent(m_hMsgAvailableEvent);
		setevent(m_hQEmptyEvent);
	}
	else
		((unsigned long*)m_pBase)[0] = (unsigned long)((LPBYTE)pHead - (LPBYTE)m_pBase);

	unlock(m_hQMutex);

	return true;
}

// Function name	: ShmemLockedQueue::RemoveNextInsert
// Description	    : 
// Return type		: bool 
// Argument         : MessageQueue *pMsgQueue
bool ShmemLockedQueue::RemoveNextInsert(MessageQueue *pMsgQueue)
{
	ShmemLockedQueueHeader *pTail, *pHead, *pMessage;

	// Get the next available entry in the queue
	while(true)
	{
		// Get the queue mutex
		lock(m_hQMutex);
		
		// Wait for the queue to not be empty
		while (((unsigned long*)m_pBase)[0] == 0)
		{
			unlock(m_hQMutex);
			//wait(m_hMsgAvailableEvent);
			if (WaitForSingleObject(m_hMsgAvailableEvent, INFINITE) != WAIT_OBJECT_0)
			{
				printf("Wait for MsgAvailableEvent failed\n");
				return false;
			}
			lock(m_hQMutex);
		}
		
		// Search the queue for the next available entry
		pMessage = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
		pTail = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
		while ((pMessage->state == SHMEM_Q_BEING_READ) && (pMessage < pTail))
			pMessage = (ShmemLockedQueueHeader*)((LPBYTE)pMessage + pMessage->next_offset);
		
		// If we haven't reached the tail and the element we are on is not currently being written, then successfully break out of this loop
		if ( (pMessage < pTail) && (pMessage->state != SHMEM_Q_BEING_WRITTEN) )
			break;

		// All messages are being read or the next message in order is not ready. I need to reset MsgAvailableEvent, wait for it to be signalled and then start over
		//resetevent(m_hMsgAvailableEvent);
		ResetEvent(m_hMsgAvailableEvent);
		unlock(m_hQMutex);
		//wait(m_hMsgAvailableEvent);
		if (WaitForSingleObject(m_hMsgAvailableEvent, INFINITE) != WAIT_OBJECT_0)
		{
			printf("Wait for MsgAvailableEvent failed\n");
			return false;
		}
	}

	MessageQueue::MsgQueueElement *pElement;
	if (pMessage->from == -1)
		return false;
	void *pBuffer = pMsgQueue->GetBuffer(pMessage->tag, pMessage->length, pMessage->from, &pElement);

	// Mark the message as being read
	pMessage->state = SHMEM_Q_BEING_READ;
	unlock(m_hQMutex);

	// Read the data from the message
	memcpy(pBuffer, (LPBYTE)pMessage + sizeof(ShmemLockedQueueHeader), pMessage->length);
	//SetEvent(pElement->hEvent);
	pMsgQueue->SetElementEvent(pElement);

	lock(m_hQMutex);
	// Mark the message as having been read
	pMessage->state = SHMEM_Q_READ;

	// Update the head and tail pointers of the queue
	pHead = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[0] );
	pTail = (ShmemLockedQueueHeader*)( (LPBYTE)m_pBase + ((unsigned long *)m_pBase)[1] );
		
	// Advance the head pointer over all the read messages
	while ((pHead->state == SHMEM_Q_READ) && (pHead < pTail))
		pHead = (ShmemLockedQueueHeader*)((LPBYTE)pHead + pHead->next_offset);

	if (pHead >= pTail)
	{
		// When the head catches up to the tail, the queue is empty so reset the pointers and signal the queue empty
		((unsigned long*)m_pBase)[0] = 0;
		((unsigned long*)m_pBase)[1] = 2*sizeof(unsigned long);
		ResetEvent(m_hMsgAvailableEvent);
		SetEvent(m_hQEmptyEvent);
	}
	else
		((unsigned long*)m_pBase)[0] = (unsigned long)((LPBYTE)pHead - (LPBYTE)m_pBase);

	unlock(m_hQMutex);

	return true;
}
