// nt_ipvishm_message_queue.h: interface for the MessageQueue class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_MESSAGEQUEUE_H__BF8A68B9_19C3_11D3_95D2_009027106653__INCLUDED_)
#define AFX_MESSAGEQUEUE_H__BF8A68B9_19C3_11D3_95D2_009027106653__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "mpidefs.h"
#include <windows.h>

// With the environment variable MPICH_USE_POLLING undefined WaitForElementEvent waits on a kernel event handle
// With MPICH_USE_POLLING defined WaitForElementEvent polls a variable in shared memory

#define MSGQ_ELEMENT_BLOCK 1000

class MessageQueue  
{
public:
	struct MsgQueueElement
	{
		HANDLE hEvent;
		BYTE cTrigger;
		bool bIn_use;
	};

	MessageQueue();
	virtual ~MessageQueue();

	void* GetBuffer(int tag, int length, int from, MsgQueueElement **ppElement);
	bool FillBuffer(int tag, void *buffer, int *length, int *from);
	bool Available(int tag, int &from);

	bool PostRecvBuffer(int tag, void *buffer, int length, int *pID);
	bool Wait(int *pID);
	bool Test(int *pID);

	inline bool SetElementEvent(MsgQueueElement *pElement);
	inline bool ResetElementEvent(MsgQueueElement *pElement);
	inline bool WaitForElementEvent(MsgQueueElement *pElement);
private:
	inline void InitElement(MsgQueueElement *pElement);
	inline void CloseElement(MsgQueueElement *pElement);
	struct InternalListNode
	{
		int length, from;
		void *buffer;
		MsgQueueElement *element;
		InternalListNode *next;
	};
	struct InternalNode
	{
		int tag;
		InternalListNode *list, *posted;
		InternalNode *left;
		InternalNode *right;
	};

	// Buffer management variables
	CRITICAL_SECTION m_CriticalSection;
	InternalNode *m_pHeap;

	// Event management variables
	int m_num_elements;
	int m_num_available;
	int m_cur_index;
	int m_pool_size;
	MsgQueueElement **m_pPool;
	bool m_bUseEvent;

	// Event management functions
	MsgQueueElement* allocElement();
	void freeElement(MsgQueueElement *element);

	// global variables used for FindNode
	int find_tag, find_length, find_from;
	void *find_buffer;
	MsgQueueElement *find_pElement;

	void FindNode(InternalNode *&node);
	bool FindAvailable(InternalNode *node, int tag, int &from);

	InternalListNode *fillfind_node;
	void FillFindNode(InternalNode *&node);

	// Cleanup functions
	void DeleteHeapList(InternalListNode *listnode);
	void DeleteHeap(InternalNode *node);
};

// Function name	: MessageQueue::SetElementEvent
// Description	    : 
// Return type		: bool 
// Argument         : MessageQueue::MsgQueueElement *pElement
inline bool MessageQueue::SetElementEvent(MessageQueue::MsgQueueElement *pElement)
{
	if (m_bUseEvent)
	{
		if (!SetEvent(pElement->hEvent))
			return false;
		return true;
	}
	pElement->cTrigger = 1;
	return true;
}

// Function name	: MessageQueue::ResetElementEvent
// Description	    : 
// Return type		: bool 
// Argument         : MessageQueue::MsgQueueElement *pElement
inline bool MessageQueue::ResetElementEvent(MessageQueue::MsgQueueElement *pElement)
{
	if (m_bUseEvent)
	{
		if (!ResetEvent(pElement->hEvent))
			return false;
		return true;
	}
	pElement->cTrigger = 0;
	return true;
}

// Function name	: MessageQueue::WaitForElementEvent
// Description	    : 
// Return type		: bool 
// Argument         : MessageQueue::MsgQueueElement *pElement
inline bool MessageQueue::WaitForElementEvent(MessageQueue::MsgQueueElement *pElement)
{
	if (m_bUseEvent)
		return (WaitForSingleObject(pElement->hEvent, INFINITE) == WAIT_OBJECT_0);
	while (pElement->cTrigger == 0)
		Sleep(0);
	return true;
}

// Function name	: MessageQueue::InitElement
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::MsgQueueElement *pElement
inline void MessageQueue::InitElement(MessageQueue::MsgQueueElement *pElement)
{
	pElement->hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	pElement->cTrigger = 0;
	pElement->bIn_use = false;
}

// Function name	: MessageQueue::CloseElement
// Description	    : 
// Return type		: void 
// Argument         : MessageQueue::MsgQueueElement *pElement
inline void MessageQueue::CloseElement(MessageQueue::MsgQueueElement *pElement)
{
	CloseHandle(pElement->hEvent);
	pElement->cTrigger = 0;
	pElement->bIn_use = false;
}

#endif // !defined(AFX_MESSAGEQUEUE_H__BF8A68B9_19C3_11D3_95D2_009027106653__INCLUDED_)
