/*
 *  $Id: inter_fns.c,v 1.3 2000/07/12 21:27:12 gropp Exp $
 *
 *  (C) 2000 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"

/* ---------------------------------
 * Modifications by Rajkumar Venkat
 */

#include "mpimem.h" 		/* Temporary buffer allocation */
#include "../coll/coll.h"
#include "mpipt2pt.h"		/* MPIR_Type_get_limits */
#include "mpiops.h"

/* End of Modifications
 * ---------------------------------
 */


#define MPI_ERR_COMM_INTER MPIR_ERRCLASS_TO_CODE(MPI_ERR_COMM,MPIR_ERR_COMM_INTER)

/* ---------------------------------
 * Modifications by Rajkumar Venkat
 */

/* Define MPI-2 attribute functions as their MPI-1 counterparts */
# define MPI_Comm_get_attr 	MPI_Attr_get
# define MPI_Comm_set_attr 	MPI_Attr_put
# define MPI_Comm_create_keyval	MPI_Keyval_create

/* This is the current assumption made
 * regarding the rank of the root/leader of
 * each of the communicators in the InterComm
 */

# define DESIGNATED_LOCAL_ROOT		0
# define DESIGNATED_REMOTE_ROOT		0

# define MPI_INTER_BCAST_PT_TAG 	25
# define MPI_INTER_REDUCE_PT_TAG	26
# define MPI_INTER_ALLREDUCE_SEND_TAG	27
# define MPI_INTER_ALLREDUCE_RECV_TAG	28

/* Used for error message construction */
# define MPI_ERR_REMOTE			29
# define MPI_ERR_REMOTE_ZERO		30

/* End of Modifications
 * ---------------------------------
 */


/*
 * Provide the collective ops structure for inter communicators.
 * Doing it this way (as a different set of functions) removes a test
 * from each collective call, and gets us back at least the cost of the
 * additional indirections and function call we have to provide the 
 * abstraction !
 *
 * Written by James Cownie (Meiko) 31 May 1995
 */

/* Forward declarations */
static int inter_Barrier (struct MPIR_COMMUNICATOR *);
static int inter_Bcast (void*, int, struct MPIR_DATATYPE *, int, struct MPIR_COMMUNICATOR *);
static int inter_Gather (void*, int, struct MPIR_DATATYPE *, void*, 
				   int, struct MPIR_DATATYPE *, int, struct MPIR_COMMUNICATOR *);
static int inter_Gatherv (void*, int, struct MPIR_DATATYPE *, void*, int *, 
				    int *, struct MPIR_DATATYPE *, int, struct MPIR_COMMUNICATOR *); 
static int inter_Scatter (void*, int, struct MPIR_DATATYPE *, void*, int, 
				    struct MPIR_DATATYPE *, int, struct MPIR_COMMUNICATOR *);
static int inter_Scatterv (void*, int *, int *, struct MPIR_DATATYPE *, 
				     void*, int, struct MPIR_DATATYPE *, int, struct MPIR_COMMUNICATOR *);
static int inter_Allgather (void*, int, struct MPIR_DATATYPE *, void*, int, 
				      struct MPIR_DATATYPE *, struct MPIR_COMMUNICATOR *);
static int inter_Allgatherv (void*, int, struct MPIR_DATATYPE *, void*, int *,
				       int *, struct MPIR_DATATYPE *, struct MPIR_COMMUNICATOR *);
static int inter_Alltoall (void*, int, struct MPIR_DATATYPE *, 
				     void*, int, struct MPIR_DATATYPE *, struct MPIR_COMMUNICATOR *);
static int inter_Alltoallv (void*, int *, int *, 
				      struct MPIR_DATATYPE *, void*, int *, 
				      int *, struct MPIR_DATATYPE *, struct MPIR_COMMUNICATOR *);
static int inter_Reduce (void*, void*, int, 
				   struct MPIR_DATATYPE *, MPI_Op, int, struct MPIR_COMMUNICATOR *);
static int inter_Allreduce (void*, void*, int, 
				      struct MPIR_DATATYPE *, MPI_Op, struct MPIR_COMMUNICATOR *);
static int inter_Reduce_scatter (void*, void*, int *, 
					   struct MPIR_DATATYPE *, MPI_Op, struct MPIR_COMMUNICATOR *);
static int inter_Scan (void*, void*, int, struct MPIR_DATATYPE *, 
				 MPI_Op, struct MPIR_COMMUNICATOR * );

static struct _MPIR_COLLOPS inter_collops = {
    inter_Barrier,
    inter_Bcast,
    inter_Gather, 
    inter_Gatherv, 
    inter_Scatter,
    inter_Scatterv,
    inter_Allgather,
    inter_Allgatherv,
    inter_Alltoall,
    inter_Alltoallv,
    0, /* inter_Alltoallw, */
    inter_Reduce,
    inter_Allreduce,
    inter_Reduce_scatter,
    inter_Scan,
    1                              /* Giving it a refcount of 1 ensures it
				    * won't ever get freed.
				    */
};

MPIR_COLLOPS MPIR_inter_collops = &inter_collops;

/* ---------------------------------
 * Modifications by Rajkumar Venkat
 */

/* Global variable for cached attributes */
int key_comm_local = MPI_KEYVAL_INVALID;

/* Internal utility function */

/* Attribute caching/retrieval...
 *   The only attribute involved is the communicator object
 *   corresponding to the local group of which the process
 *   is a member.
 * A separate function that all inter_* methods can invoke
 */
void inter_attr_prepare(MPI_Comm *comm_local,	/* 'out' parameter */
				struct MPIR_COMMUNICATOR * comm_coll)
{
	int flag;

	struct MPIR_COMMUNICATOR *comm_local_ptr;
	
	/* 'comm_coll' is assumed to represent a valid communicator
	 * So, no checking. But it could be a NULL communicator.
	 * The attribute caching is done anyway. Is it correct?
	 */

	if(key_comm_local == MPI_KEYVAL_INVALID)
	{
		/* Create a keyval for holding the local Comm */
		MPI_Comm_create_keyval(MPI_COMM_NULL_COPY_FN,
					MPI_COMM_NULL_DELETE_FN,
					&key_comm_local,
					 (void *) NULL);	/* extra_state is not used */

		/* Now create a communicator for Local Group */
		MPI_Comm_create(comm_coll->self, comm_coll->local_group->self, comm_local);

		comm_local_ptr = MPIR_GET_COMM_PTR(*comm_local);

		/* Cache the newly created local Comm as attribute of the InterComm */
		MPI_Comm_set_attr(comm_coll->self, key_comm_local, comm_local_ptr);
	}
	else
	{
		/* Get the cached attribute */
		MPI_Comm_get_attr(comm_coll->self, key_comm_local, &comm_local_ptr, &flag);
		*comm_local = comm_local_ptr->self;
	}

	/* The attribute is ready to be used at this stage */
}

/* End of Modifications
 * ---------------------------------
 */

/* Now the functions, each one simply raises an error */

static int inter_Barrier( struct MPIR_COMMUNICATOR * comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_BARRIER");
}

static int inter_Bcast ( 
	void *buffer, 
	int count, 
	struct MPIR_DATATYPE *datatype, 
	int root, 
	struct MPIR_COMMUNICATOR *comm )
{

/* ---------------------------------
 * Modifications by Rajkumar Venkat
 */

	/* Communicator Attributes */
	MPI_Comm comm_local, comm_intra;

	/* Temporary variables */
	MPI_Status status;
	MPI_Group intra_group;

	int my_rank, dest_rank, remote_root = DESIGNATED_REMOTE_ROOT;
	int comm_size, remote_size, mpi_errno = MPI_SUCCESS;
	static char myname[] = "MPI_BCAST";

	/* Switch to hidden collective */
	struct MPIR_COMMUNICATOR *comm_coll = comm->comm_coll;	
	
	/* comm = comm->comm_coll; (Isn't the above st. better?) */

	/* Check for zero-length message
	 * To verify if it is not allowed or not
	 */
	if(count == 0)
		return MPI_SUCCESS;

	/* Since we are here, the local communicator is obviously
	 * NOT NULL. However, the remote group could be NULL i.e.
	 * with no processes at all! So we check the remote size
	 * if it is zero, bail out! Set the appropriate error msg
	 * Should I be returning MPI_SUCCESS?
	 */
	remote_size = comm_coll->np;

	/* Conditional checking? */
#ifndef MPIR_NO_ERROR_CHECKING
	if(remote_size <= 0)
	{
		/* To verify if this is OK...
		 * New #define's need to be added for remote group being
		 * empty. Suggestions are already included in the code below.
		 */
		mpi_errno = MPIR_Err_setmsg(MPI_ERR_REMOTE, MPIR_ERR_REMOTE_ZERO,
                                   myname, (char *)0, (char *)0, root, remote_size);
		if (mpi_errno)
			return MPIR_ERROR(comm_coll, mpi_errno, myname);
	}
#endif

	/* Get/create the cached local communicator */
	inter_attr_prepare(&comm_local, comm_coll);
	MPI_Comm_size(comm_local, &comm_size);
	MPI_Comm_rank(comm_local, &my_rank);

	/* Get the underlying "safe" intraComm
	 * from the "safe" interComm. Also get its group.
	 * NOTE: Be sure to test the validity of the
	 * underlying intraComm
	 */
	comm_intra = comm_coll->comm_coll->self;
	MPI_Comm_group(comm_intra, &intra_group);

	/* If required, insert the check for the rank of the root
	 * to see if it does not exceed the # of processes itself.
	 */

	/* Lock for collective operation */
	MPID_THREAD_LOCK(comm_coll->ADIctx, comm_coll);

	/* -- Algorithm for inter_Bcast --
	 * Check if the rank of the root...
	 *  - If it is positive, I am in the receiving group
	 *    . If I am the designated root, I receive the data
	 *      from the "root" on the other side (actually by
	 *      translating the rank of "root" to its rank in the
	 *      underlying intraComm) and participate in the broadcast
	 *      in my local group
         *    . Otherwise, simply participate in the broadcast
	 *  - If it is MPI_ROOT, then I am the root in the
	 *    sending group for this broadcast
	 *    . Send the data to the designated remote root (actually by
	 *      translating the rank of remote root to its rank in the
	 *      underlying intraComm)
	 *  - If it is MPI_PROC_NULL, then I am one of the
	 *    "uninvolved" processes of the sending group
	 *    (NOTE: I could also happen to be the designated root!)
	 *    . So, do nothing!
	 */

	if(root == MPI_ROOT)
	{
		/* I AM the source of this broadcast...
		 * So, send the data to the designated remote root
		 */
		MPI_Group_translate_ranks(comm_coll->group->self, 1, &remote_root, 
						intra_group, &dest_rank);

		mpi_errno = MPI_Send(buffer, count, datatype->self, dest_rank,
					MPI_INTER_BCAST_PT_TAG, comm_intra);
	}
	else if(root != MPI_PROC_NULL)
	{
		/* Remote Group i.e. Receiving Group */
		if(my_rank == DESIGNATED_REMOTE_ROOT)
		{
			/* Receive the data from the actual root */
			MPI_Group_translate_ranks(comm_coll->group->self, 1, &root,
							intra_group, &dest_rank);

			mpi_errno = MPI_Recv(buffer, count, datatype->self, dest_rank,
					MPI_INTER_BCAST_PT_TAG, comm_intra, &status);
		}

		/* Now, do a broadcast within my local group
		 * Before that, check if this process is the only process
		 * in the group. If so, we can skip this Bcast call itself
		 */

		/* Note that the following Bcast should invoke
		 * intra_Bcast() 
		 */
		if(comm_size > 1)
			mpi_errno = MPI_Bcast(buffer, count, datatype->self,
						 DESIGNATED_REMOTE_ROOT, comm_local);
	}

	/* Unlock to end collective operation */
	MPID_THREAD_UNLOCK(comm_coll->ADIctx, comm_coll);

	/* Some observations/questions...
	 * 1. We need to define MPI_INTER_BCAST_TAG in the list of tags. Should the
	 *    naming format be changed? In which file should such definitions appear?
	 * 2. In switching to hidden collective, a new pointer variable is
	 *    used. However, in the intra_Bcast implementation, the
 	 *    parameter "comm" itself is switched. Will it not affect the
	 *    value of "comm" for subsequent calls?
	 *    Ans. Doesn't affect
	 */

	return mpi_errno;

/* End of Modifications
 * ---------------------------------
 */
	
}

static int inter_Gather ( 
	void *sendbuf, 
	int sendcnt, 
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int recvcount, 
	struct MPIR_DATATYPE *recvtype, 
	int root, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_GATHER");
}

static int inter_Gatherv ( 
	void *sendbuf, 
	int sendcnt,  
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int *recvcnts, 
	int *displs, 
	struct MPIR_DATATYPE *recvtype, 
	int root, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_GATHERV");
}

static int inter_Scatter ( 
	void *sendbuf, 
	int sendcnt, 
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int recvcnt, 
	struct MPIR_DATATYPE *recvtype, 
	int root, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_SCATTER");
}

static int inter_Scatterv ( 
	void *sendbuf, 
	int *sendcnts, 
	int *displs, 
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int recvcnt,  
	struct MPIR_DATATYPE *recvtype, 
	int root, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_SCATTERV");
}

static int inter_Allgather ( 
	void *sendbuf, 
	int sendcount, 
	struct MPIR_DATATYPE *sendtype,
	void *recvbuf, 
	int recvcount, 
	struct MPIR_DATATYPE *recvtype, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_ALLGATHER");
}

static int inter_Allgatherv ( 
	void *sendbuf, 
	int sendcount,  
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int *recvcounts, 
	int *displs,   
	struct MPIR_DATATYPE *recvtype, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_ALLGATHERV");
}

static int inter_Alltoall( 
	void *sendbuf, 
	int sendcount, 
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int recvcnt, 
	struct MPIR_DATATYPE *recvtype, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_GATHERV");
}

static int inter_Alltoallv ( 
	void *sendbuf, 
	int *sendcnts, 
	int *sdispls, 
	struct MPIR_DATATYPE *sendtype, 
	void *recvbuf, 
	int *recvcnts, 
	int *rdispls, 
	struct MPIR_DATATYPE *recvtype, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_ALLTOALLV");
}

static int inter_Reduce ( 
	void *sendbuf, 
	void *recvbuf, 
	int count, 
	struct MPIR_DATATYPE *datatype, 
	MPI_Op op, 
	int root, 
	struct MPIR_COMMUNICATOR *comm )
{

/* ---------------------------------
 * Modifications by Rajkumar Venkat
 */

	/* Communicator Attributes */
	MPI_Comm comm_local, comm_intra;

	/* Temporary variables */
	MPI_Aint ub, lb, m_extent;	/* Buffer Allocation */
	MPI_Status status;
	MPI_Group intra_group;

	int my_rank, dest_rank, remote_root = DESIGNATED_REMOTE_ROOT;
	int comm_size, remote_size, mpi_errno = MPI_SUCCESS;
	void *buffer;
	static char myname[] = "MPI_REDUCE";

	/* Switch to hidden collective */
	struct MPIR_COMMUNICATOR *comm_coll = comm->comm_coll;

	/* comm = comm->comm_coll; */

	/* Check for zero-length message
	 * To verify if it is allowed or not
	 */
	if(count == 0)
		return MPI_SUCCESS;

	/* Since we are here, the local communicator is obviously
	 * NOT NULL. However, the remote group could be NULL i.e.
	 * with no processes at all! So we check the remote size
	 * if it is zero, bail out! Set the appropriate error msg
	 */
	remote_size = comm_coll->np;

	/* Conditional checking? */
#ifndef MPIR_NO_ERROR_CHECKING
	if(remote_size <= 0)
	{
		/* To verify if this is OK...
		 * New #define's need to be added for remote group being
		 * empty. Suggestions are already included in the code below.
		 */
		mpi_errno = MPIR_Err_setmsg(MPI_ERR_REMOTE, MPIR_ERR_REMOTE_ZERO,
                                   myname, (char *)0, (char *)0, root, remote_size);
		if (mpi_errno)
			return MPIR_ERROR(comm_coll, mpi_errno, myname);
	}
#endif

	/* Additional Reduce-specific checks include checking if
	 * the 'Op' is valid for the 'datatype', if predefined.
	 * As these checks are done in the intra_Reduce, I need
	 * not check it here. Correct?
	 */

	/* Get the cached local communicator */
	inter_attr_prepare(&comm_local, comm_coll);
	MPI_Comm_size(comm_local, &comm_size);
	MPI_Comm_rank(comm_local, &my_rank);

	/* Get the underlying "safe" intraComm
	 * from the "safe" interComm. Also get its group.
	 * NOTE: Be sure to test the validity of the
	 * underlying intraComm
	 */
	comm_intra = comm_coll->comm_coll->self;
	MPI_Comm_group(comm_intra, &intra_group);

	/* If required, insert the check for the rank of the root
	 * to see if it does not exceed the # of processes itself.
	 */

	if((root >= 0) && (my_rank == DESIGNATED_REMOTE_ROOT) && (comm_size > 1))
	{
		/* I'm the remote root and the root for
		 * the intra-reduce... 
		 * Allocate a temporary extra buffer
		 */
 		MPIR_Type_get_limits(datatype, &lb, &ub);
  		m_extent = ub - lb;

		/* The reason for using a different variable
		 * for the buffer instead of using the recvbuf
		 * is: As the recvbuf is insignificant except
		 * at the root, the recvbuf passed could be
		 * "anything" - even a valid buffer (allocated
		 * in the calling program for e.g.) and expect
		 * it to remain intact at the end of this call
		 * as it is supposed to be ignored by the
		 * MPI_Reduce call. Actually, intra_Reduce reuses
		 * the variable currently. Is it Ok?
		 */
 
		MPIR_ALLOC(buffer,
				(void *)MALLOC(m_extent * count),
				comm_local, MPI_ERR_EXHAUSTED,
           			"MPI_REDUCE" );
		buffer = (void *)((char *)buffer - lb);
        }

	/* Lock for collective operation */ 
	MPID_THREAD_LOCK(comm_coll->ADIctx, comm_coll);
                
	/* -- Algorithm for inter_Reduce --
	 * - Perform MPI_Reduce (intra_reduce) on the "remote"
	 *   group rooted at the designated "remote" root
	 *   (Note that we DO need to have an additional buffer
	 *   at the designated "remote" root to hold the result).
	 * - Send the result to the actual root of this operation
	 *   in the "local" group through the underlying intra-comm
	 */

	if(root == MPI_ROOT)
	{
		/* I'm the root for this Reduce operation
		 * So, I just receive the "reduced" data
		 * from the "remote" root
		 */
		MPI_Group_translate_ranks(comm_coll->group->self,
						1, &remote_root,
						intra_group, &dest_rank); 

		mpi_errno = MPI_Recv(recvbuf, count,
					datatype->self, dest_rank,
					MPI_INTER_REDUCE_PT_TAG,
					comm_intra, &status);
	}
	else if(root != MPI_PROC_NULL)
	{
		/* I am a process in the remote group */

		/* Perform an intra_reduce...
		 * The following MPI_Reduce should invoke
	 	 * the corresponding "intra" version  
	 	 */
		if(comm_size > 1)
		{
			MPI_Reduce(sendbuf, buffer, count, datatype->self,
				 op, DESIGNATED_REMOTE_ROOT, comm_local);
		}

		if(my_rank == DESIGNATED_REMOTE_ROOT)
		{
			/* Send the "reduced" data to the other side */
			MPI_Group_translate_ranks(comm_coll->group->self,
						1, &root, intra_group,
						&dest_rank); 

			if(comm_size > 1)
			{
				mpi_errno = MPI_Send(buffer, count,
						datatype->self, dest_rank,
						MPI_INTER_REDUCE_PT_TAG,
						comm_intra);
				/* Now that the work is over,
				 * free the buffer
				 * Again, find out why it is
				 * not just FREE(buffer)...
				 */
				FREE((char *)buffer + lb);	
			}
			else
			{
				mpi_errno = MPI_Send(sendbuf, count,
						datatype->self, dest_rank,
						MPI_INTER_REDUCE_PT_TAG,
						comm_intra);
			}

		}
	}

        /* Unlock to end collective operation */
        MPID_THREAD_UNLOCK(comm_coll->ADIctx, comm_coll);

        /* Some observations/questions...   
         * 1. We need to define MPI_INTER_REDUCE_PT_TAG in the list of tags. Should
         *    the naming format be changed? In which file should such definitions
	 *    appear?
	 * 2. What about buffer allocation and de-allocation failures? They
	 *    are not taken care of in intra_Reduce as well.
	 *    Ans. Don't worry about it for now
         */
                                           
        return mpi_errno;
                                             
/* End of Modifications
 * ---------------------------------
 */

}

static int inter_Allreduce ( 
	void *sendbuf, 
	void *recvbuf, 
	int count, 
	struct MPIR_DATATYPE *datatype, 
	MPI_Op op, 
	struct MPIR_COMMUNICATOR *comm )
{

/* ---------------------------------
 * Modifications by Rajkumar Venkat  
 */
	/* Communicator Attributes */
	MPI_Comm comm_local, comm_intra;
        
	/* Temporary variables */
	MPI_Status status;
	MPI_Group intra_group;
 
        int my_rank, dest_rank, remote_root = DESIGNATED_REMOTE_ROOT;
        int comm_size, remote_size, mpi_errno = MPI_SUCCESS;
        static char myname[] = "MPI_ALLREDUCE";
        
        /* Switch to hidden collective */
        struct MPIR_COMMUNICATOR *comm_coll = comm->comm_coll;
        
        /* comm = comm->comm_coll; */
        
        /* Check for zero-length message
         * To verify if it is allowed or not
         */
        if(count == 0)
                return MPI_SUCCESS;
        
	/* Since we are here, the local communicator is obviously
	 * NOT NULL. However, the remote group could be NULL i.e.
	 * with no processes at all! So we check the remote size
	 * if it is zero, bail out! Set the appropriate error msg
	 */
	remote_size = comm_coll->np;
	/* Conditional checking? */
#ifndef MPIR_NO_ERROR_CHECKING
	if(remote_size <= 0)
	{
		/* To verify if this is OK...
		 * New #define's need to be added for remote group being
		 * empty. Suggestions are already included in the code below.
		 */
		mpi_errno = MPIR_Err_setmsg(MPI_ERR_REMOTE, MPIR_ERR_REMOTE_ZERO,
                                   myname, (char *)0, (char *)0, root, remote_size);
		if (mpi_errno)
			return MPIR_ERROR(comm_coll, mpi_errno, myname);
	}
#endif

	/* Additional Reduce-specific checks include checking if
	 * the 'Op' is valid for the 'datatype', if predefined.
	 * As these checks are done in the intra_Reduce, I need
	 * not check it here. Correct?
	 */

	/* Get the cached local communicator */
	inter_attr_prepare(&comm_local, comm_coll);
	MPI_Comm_size(comm_local, &comm_size);
	MPI_Comm_rank(comm_local, &my_rank);

	/* Get the underlying "safe" intraComm
	 * from the "safe" interComm. Also get its group.
	 * NOTE: Be sure to test the validity of the
	 * underlying intraComm
	 */
	comm_intra = comm_coll->comm_coll->self;
	MPI_Comm_group(comm_intra, &intra_group);
         
        /* If required, insert the check for the rank of the root
         * to see if it does not exceed the # of processes itself.
         */
         
        /* Lock for collective operation */
        MPID_THREAD_LOCK(comm_coll->ADIctx,comm_coll);

        /* -- Algorithm for inter_Allreduce --
	 * - Perform MPI_Reduce (intra_reduce) on the respective
	 *   groups with the same sendbuf and rooted at the
	 *   designated "local" root for each group. Note that
	 *   all the processes OTHER THAN the root can pass in the
	 *   same recvbuf, but with no effect, as the recvbuf will
	 *   be significant only at the designated root.
	 * - At this point, the designated roots have the result of
	 *   the reduction of their local groups. Just send this result
	 *   over to the designated root on the other side using
	 *   MPI_Sendrecv(). Apparently, there seems to be NO issue
	 *   with mismatch of the send and recv buffers as the count
	 *   and datatype are common. Further, the problem of checking
	 *   if the data is contiguous or not rests with the underlying
	 *   implementation of the intra_Allreduce, NOT here (?).
	 * - Now the results are in the correct groups - but just at the
	 *   designated root. So locally do an intra_Bcast()
	 */

	/* The following MPI_Reduce and MPI_Bcast should invoke
	 * the corresponding "intra" versions
	 */
	if(comm_size > 1)
	{
		mpi_errno = MPI_Reduce(sendbuf, recvbuf,
					count, datatype->self, op,
					DESIGNATED_LOCAL_ROOT,
					comm_local);
	}
	if(my_rank == DESIGNATED_LOCAL_ROOT)
	{
		/* Send to AND get from the other side */
		MPI_Group_translate_ranks(comm_coll->group->self,
					   1, &remote_root,
					    intra_group, &dest_rank);

		mpi_errno = MPI_Sendrecv(recvbuf, count, datatype->self,
					  dest_rank, MPI_INTER_ALLREDUCE_SEND_TAG,
					   sendbuf, count, datatype->self,
					    dest_rank, MPI_INTER_ALLREDUCE_RECV_TAG,
					     comm_intra, &status);
	}		
	if(comm_size > 1)
	{
		mpi_errno = MPI_Bcast(sendbuf, count, datatype->self,
					 DESIGNATED_LOCAL_ROOT, comm_local);
	}
                 
        /* Unlock to end collective operation */
        MPID_THREAD_UNLOCK(comm_coll->ADIctx,comm_coll);

        /* Some observations/questions...
         * 1. We need to define MPI_INTER_ALLREDUCE_SEND_TAG
	 *    and MPI_INTER_ALLREDUCE_RECV_TAG in the list of tags. Should the
         *    naming format be changed? In which file should such definitions
	 *    appear?
	 * 2. How come the intra_Allreduce looks so simple???
	 *    Ans. Because the underlying calls do ALL the checking.
         */      
        
        return mpi_errno;
                
/* End of Modifications
 * ---------------------------------
 */

}

static int inter_Reduce_scatter ( 
	void *sendbuf, 
	void *recvbuf, 
	int *recvcnts, 
	struct MPIR_DATATYPE *datatype, 
	MPI_Op op, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_REDUCE_SCATTER");
}

static int inter_Scan ( 
	void *sendbuf, 
	void *recvbuf, 
	int count, 
	struct MPIR_DATATYPE *datatype, 
	MPI_Op op, 
	struct MPIR_COMMUNICATOR *comm )
{
    return MPIR_ERROR(comm, MPI_ERR_COMM_INTER,
		      "MPI_SCAN");
}

